"""Constants for the data module."""

from napistu_torch.constants import (
    MASK_HASHES,
    NAPISTU_DATA,
    NAPISTU_DATA_SUMMARIES,
    PYG,
)

DEFAULT_SAME_DATA_ALLOW_MISSING_KEYS = {
    PYG.NUM_EDGE_FEATURES,
    NAPISTU_DATA_SUMMARIES.NUM_UNIQUE_RELATIONS,
    NAPISTU_DATA_SUMMARIES.NUM_UNIQUE_CLASSES,
    NAPISTU_DATA.RELATION_TYPE_LABELS,
} | set(MASK_HASHES)
